#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include "config.h"
#include "charset.h"

#define KOI8R		0
#define CP1251		1
#define	CP866		2
#define	ISO88595	3
#define MACCYR		4
#define CP1250		5
#define ISO88591	6
#define ISO88592	7

typedef	unsigned char TABLE[256];
typedef struct charset_struct {
	int		charset;
	unsigned	char *chars;
	TABLE		lower;
	TABLE		upper;
} TCHARSET;

typedef struct alias_struct {
	int		charset;
	unsigned	char *alias;
} ALIAS;



/* cyrillic charsets */
static unsigned char cp1251[]  = "Ũ";
static unsigned char koi8r[]   = "ţ";
static unsigned char cp866[]   = "񦧨";
static unsigned char iso88595[]= "";
static unsigned char maccyr[]  = "݆";

/* Western */
static unsigned char iso88591[] ="";

/* Central europe  */
static unsigned char iso88592[] = "߱";
static unsigned char cp1250[]   = "߹";

#define MAX_CHARSET	8
static TCHARSET Charsets[MAX_CHARSET] ={
	{KOI8R,koi8r,""},{CP1251,cp1251,""},{CP866,cp866,""},
	{ISO88595,iso88595,""},{MACCYR,maccyr,""},
	{CP1250,cp1250,""},{ISO88591,iso88591,""},{ISO88592,iso88592,""}
};


#define MAX_ALIAS	50
static ALIAS Aliases[50] = {
	/* Cyrillic */
	{KOI8R,"koi8r"},{KOI8R,"koi8-r"},{KOI8R,"koi-8-r"},
	{KOI8R,"koi8"},	{KOI8R,"koi-8"},{KOI8R,"koi"},
	{CP1251,"windows-1251"},{CP1251,"windows1251"},{CP1251,"win"},
	{CP1251,"x-cp1251"},{CP1251,"cp1251"},{CP1251,"cp-1251"},
	{ISO88595,"ISO-8859-5"},{ISO88595,"iso_8859-5:1988"},
	{ISO88595,"iso-ir-144"},{ISO88595,"iso_8859-5"},
	{ISO88595,"cyrillic"},{ISO88595,"ISO-8859-5"},{ISO88595,"iso8859-5"},
	{ISO88595,"iso-8859.5"},{ISO88595,"iso8859.5"},
	{CP866,"ibm866"},{CP866,"csibm866"},{CP866,"866"},{CP866,"cp866"},
	{CP866,"x-cp866"},{CP866,"x-ibm866"},{CP866,"cp-866"},{CP866,"alt"},
	{MACCYR,"x-mac-cyrillic"},
	/* cp1250 */
	{CP1250,"windows-1250"},{CP1250,"windows1250"},{CP1250,"win"},
	{CP1250,"x-cp1250"},{CP1250,"cp1250"},{CP1250,"cp-1250"},
	/* 8859-1 */
	{ISO88591,"ISO-8859-1"},{ISO88591,"iso_8859-1:1988"},
	{ISO88591,"iso_8859-1"},{ISO88591,"ISO-8859-1"},
	{ISO88591,"iso8859-1"},{ISO88591,"iso-8859.1"},{ISO88591,"iso8859.1"},
	/* 8859-2 */
	{ISO88592,"ISO-8859-2"},{ISO88592,"iso_8859-2:1988"},
	{ISO88592,"iso_8859-2"},{ISO88592,"ISO-8859-2"},
	{ISO88592,"iso8859-2"},{ISO88592,"iso-8859.2"},{ISO88592,"iso8859.2"}
};

/*
typedef struct recode_struce {
	int		charset_from;
	int		charset_to;
} RECODE_PAIRS;

#define MAX_RECODE 18
static RECODE_PAIRS RecodePairs[MAX_RECODE]={
	{KOI8R,CP1251},{KOI8R,CP866},{KOI8R,MACCYR},{KOI8R,ISO88595},
	{CP1251,KOI8R},{CP1251,CP866},{CP1251,MACCYR},{CP1251,ISO88595},
	{CP866,CP1251},{CP866,KOI8R},{CP866,MACCYR},{CP866,ISO88595},
	{ISO88595,CP1251},{ISO88595,KOI8R},{ISO88595,MACCYR},{ISO88595,CP866},
	{ISO88592,CP1250},{CP1250,ISO88592}
};
*/

char SEPARATOR[256];

void PrintTable(TABLE t){
int i;
	for(i=32;i<256;i++){
		if(!(i&31))printf("\n");
		printf("%c",t[i]);
	}
}


int getcharset(unsigned char *alias){
int i;
	if(!alias)return(MAX_CHARSET);
	for(i=0;i<MAX_ALIAS;i++)
		if(!strcasecmp(Aliases[i].alias,alias))
			return(Aliases[i].charset);
	return(MAX_CHARSET);
}


char *tolower_string(unsigned char *str, int charset){
unsigned char *s;
TABLE *t;

	if(!(s=str))return(NULL);
	if(charset<MAX_CHARSET){
		t=&(Charsets[charset].lower);
		s=str;
		while(*s){
			*s=*(*t+*s);
			s++;
		}
	}else{
		while(*s){
			*s=tolower(*s);
			s++;
		}
	}
	return(str);
}


char *toupper_string(unsigned char *str, int charset){
unsigned char *s;
TABLE *t;

	if(!(s=str))return(NULL);
	if(charset<MAX_CHARSET){
		t=&(Charsets[charset].upper);
		s=str;
		while(*s){
			*s=*(*t+*s);
			s++;
		}
	}else{
		while(*s){
			*s=toupper(*s);
			s++;
		}
	}
	return(str);
}


#define SGML_CHAR_NUM 101
static struct
{
    char		*sgml;
    unsigned char	equiv;
} SGMLChars[SGML_CHAR_NUM] =
  {
    { "lt",	      '<' } ,
    { "gt",	      '>' } ,
    { "amp",	      '&' } ,
    { "quot",	      '"' } ,
    { "nbsp",         ' ' } , /* non breaking space */

    { "trade",	      153 } , /* trade mark */
    { "iexcl",        161 } , /* inverted exclamation mark */
    { "cent",         162 } , /* cent sign */
    { "pound",        163 } , /* pound sign */
    { "curren",       164 } , /* currency sign */
    { "yen",          165 } , /* yen sign */
    { "brvbar",       166 } , /* broken vertical bar, (brkbar) */
    { "sect",         167 } , /* section sign */
    { "uml",          168 } , /* spacing diaresis */
    { "copy",         169 } , /* copyright sign */
    { "ordf",         170 } , /* feminine ordinal indicator */
    { "laquo",        171 } , /* angle quotation mark, left */
    { "not",          172 } , /* negation sign */
    { "shy",          173 } , /* soft hyphen */
    { "reg",          174 } , /* circled R registered sign */
    { "hibar",        175 } , /* spacing macron */
    { "deg",          176 } , /* degree sign */
    { "plusmn",       177 } , /* plus-or-minus sign */
    { "sup2",         178 } , /* superscript 2 */
    { "sup3",         179 } , /* superscript 3 */
    { "acute",        180 } , /* spacing acute (96) */
    { "micro",        181 } , /* micro sign */
    { "para",         182 } , /* paragraph sign */
    { "middot",       183 } , /* middle dot */
    { "cedil",        184 } , /* spacing cedilla */
    { "sup1",         185 } , /* superscript 1 */
    { "ordm",         186 } , /* masculine ordinal indicator */
    { "raquo",        187 } , /* angle quotation mark, right */
    { "frac14",       188 } , /* fraction 1/4 */
    { "frac12",       189 } , /* fraction 1/2 */
    { "frac34",       190 } , /* fraction 3/4 */
    { "iquest",       191 } , /* inverted question mark */
    { "Agrave",       192 } , /* capital A, grave accent */ 
    { "Aacute",       193 } , /* capital A, acute accent */ 
    { "Acirc",        194 } , /* capital A, circumflex accent */ 
    { "Atilde",       195 } , /* capital A, tilde */ 
    { "Auml",         196 } , /* capital A, dieresis or umlaut mark */ 
    { "Aring",        197 } , /* capital A, ring */ 
    { "AElig",        198 } , /* capital AE diphthong (ligature) */ 
    { "Ccedil",       199 } , /* capital C, cedilla */ 
    { "Egrave",       200 } , /* capital E, grave accent */ 
    { "Eacute",       201 } , /* capital E, acute accent */ 
    { "Ecirc",        202 } , /* capital E, circumflex accent */ 
    { "Euml",         203 } , /* capital E, dieresis or umlaut mark */ 
    { "Igrave",       205 } , /* capital I, grave accent */ 
    { "Iacute",       204 } , /* capital I, acute accent */ 
    { "Icirc",        206 } , /* capital I, circumflex accent */ 
    { "Iuml",         207 } , /* capital I, dieresis or umlaut mark */ 
    { "ETH",          208 } , /* capital Eth, Icelandic (Dstrok) */ 
    { "Ntilde",       209 } , /* capital N, tilde */ 
    { "Ograve",       210 } , /* capital O, grave accent */ 
    { "Oacute",       211 } , /* capital O, acute accent */ 
    { "Ocirc",        212 } , /* capital O, circumflex accent */ 
    { "Otilde",       213 } , /* capital O, tilde */ 
    { "Ouml",         214 } , /* capital O, dieresis or umlaut mark */ 
    { "times",        215 } , /* multiplication sign */ 
    { "Oslash",       216 } , /* capital O, slash */ 
    { "Ugrave",       217 } , /* capital U, grave accent */ 
    { "Uacute",       218 } , /* capital U, acute accent */ 
    { "Ucirc",        219 } , /* capital U, circumflex accent */ 
    { "Uuml",         220 } , /* capital U, dieresis or umlaut mark */ 
    { "Yacute",       221 } , /* capital Y, acute accent */ 
    { "THORN",        222 } , /* capital THORN, Icelandic */ 
    { "szlig",        223 } , /* small sharp s, German (sz ligature) */ 
    { "agrave",       224 } , /* small a, grave accent */ 
    { "aacute",       225 } , /* small a, acute accent */ 
    { "acirc",        226 } , /* small a, circumflex accent */ 
    { "atilde",       227 } , /* small a, tilde */
    { "auml",         228 } , /* small a, dieresis or umlaut mark */ 
    { "aring",        229 } , /* small a, ring */
    { "aelig",        230 } , /* small ae diphthong (ligature) */ 
    { "ccedil",       231 } , /* small c, cedilla */ 
    { "egrave",       232 } , /* small e, grave accent */ 
    { "eacute",       233 } , /* small e, acute accent */ 
    { "ecirc",        234 } , /* small e, circumflex accent */ 
    { "euml",         235 } , /* small e, dieresis or umlaut mark */ 
    { "igrave",       236 } , /* small i, grave accent */ 
    { "iacute",       237 } , /* small i, acute accent */ 
    { "icirc",        238 } , /* small i, circumflex accent */ 
    { "iuml",         239 } , /* small i, dieresis or umlaut mark */ 
    { "eth",          240 } , /* small eth, Icelandic */ 
    { "ntilde",       241 } , /* small n, tilde */ 
    { "ograve",       242 } , /* small o, grave accent */ 
    { "oacute",       243 } , /* small o, acute accent */ 
    { "ocirc",        244 } , /* small o, circumflex accent */ 
    { "otilde",       245 } , /* small o, tilde */ 
    { "ouml",         246 } , /* small o, dieresis or umlaut mark */ 
    { "divide",       247 } , /* division sign */
    { "oslash",       248 } , /* small o, slash */ 
    { "ugrave",       249 } , /* small u, grave accent */ 
    { "uacute",       250 } , /* small u, acute accent */ 
    { "ucirc",        251 } , /* small u, circumflex accent */ 
    { "uuml",         252 } , /* small u, dieresis or umlaut mark */ 
    { "yacute",       253 } , /* small y, acute accent */ 
    { "thorn",        254 } , /* small thorn, Icelandic */ 
    { "yuml",         255 }  /* small y, dieresis or umlaut mark */
  };


char SgmlToChar(char *sgml,int charset){
int i;
unsigned char res[2];

	for(i=0;i<SGML_CHAR_NUM;i++)
		if(!strncasecmp(sgml,SGMLChars[i].sgml,
				strlen(SGMLChars[i].sgml))){

			res[0]=SGMLChars[i].equiv;res[1]=0;

			if(islower(*(unsigned char*)(sgml))){
				tolower_string(res,charset);
			}else{
				toupper_string(res,charset);
			}
			return((char)res[0]);
		}
	return(0);
}


static int FillLowerString(unsigned char *str,TABLE table){
unsigned int i,len;

	len=strlen(str)/2;
	for(i=0;i<256;i++)table[i]=tolower((unsigned char)i);
	for(i=0;i<len;i++){
		table[str[i]]=str[i+len];
	}
	return(0);
}

static int FillUpperString(unsigned char *str,TABLE table){
unsigned int i,len;

	len=strlen(str)/2;
	for(i=0;i<256;i++)table[i]=toupper((unsigned char)i);
	for(i=0;i<len;i++){
		table[str[i+len]]=str[i];
	}
	return(0);
}

static int FillRecodeString(unsigned char *from,unsigned char *to,TABLE table){
unsigned int len,i;

	for(i=0;i<256;i++)table[i]=(unsigned char)i;
	len=strlen(from);
	if(len!=strlen(to))return(1);
	for(i=0;i<len;i++){
		if(from[i]!=to[i])
			table[(unsigned int)from[i]]=to[i];
	}
	return(0);
}


char *recode_string(unsigned char *str,int from,int to){
unsigned char *s;
TABLE t;
	if(!(s=str))return(0);
	if(from>=MAX_CHARSET)return(0);
	if(to>=MAX_CHARSET)return(0);
	if(from==to)return(0);
	FillRecodeString(Charsets[from].chars,Charsets[to].chars,t);
	while(*s){
		*s=*(t+*s);
		s++;
	}
	return(str);
}

int init_charset(){
int i;
	for(i=0;i<MAX_CHARSET;i++){
		FillUpperString(Charsets[i].chars,Charsets[i].upper);
		FillLowerString(Charsets[i].chars,Charsets[i].lower);
	}
	return(0);
}

unsigned char *makewordchar(unsigned char * dst,int charset){
	strcpy(dst,WORDCHAR);
	if(charset<MAX_CHARSET)
		strcat(dst,Charsets[charset].chars);
	return(0);
}
