#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/time.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <regex.h>
#include <signal.h>
#ifdef HAVE_LIBUTIL_H
#include <libutil.h>
#endif
#ifdef HAVE_GETOPT_H
#include <getopt.h>
#endif
#include "common.h"
#include "log.h"
#include "indexer.h"
#include "dbms.h"
#include "parseurl.h"
#include "charset.h"
#include "parser.h"
#include "spell.h"
#include "md5.h"
#include "proto.h"
#define MD5_DIGEST(data, size, digest) md5_buffer_str(data, size, digest)
#ifndef CONFIGDIR
#define CONFIGDIR "."
#endif

#define HTML_UNK 0
#define HTML_TAG 1
#define HTML_TXT 2
#define HTML_COM 3

char *localcharset= NULL;
int lcindex=10000;
extern int loglevel;

#ifdef HAVE_SYSLOG_H
int log_facility=LOG_FACILITY;
#else
int log_facility=0;
#endif

int wcur=0,wsize=0;
UDM_WORD *Word=NULL;

int hcur=0,hsize=0;
HREF *Href=NULL;

int nservers=0;
int mservers=0;
SERVER *Server=NULL;
ROBOT *Robots=NULL;

typedef struct filter_struct {
	regex_t filter;
	int     filter_type;
	int	reverse;
	char	*regstr;
} UDM_FILTER;
int nfilters=0;
UDM_FILTER Filter[MAXFILTER];

typedef struct alias_struct {
	char *find;
	char *replace;
} ALIAS;
int naliases=0;
ALIAS Alias[MAXFILTER];


#ifndef HAVE_TM_GMTOFF
int get_gmtoff()
{
time_t tt = time(NULL);
struct tm gmt;
struct tm *t;
int days, hours, minutes;

	gmt = *gmtime(&tt);
	t = localtime(&tt);
	days = t->tm_yday - gmt.tm_yday;
	hours = ((days < -1 ? 24 : 1 < days ? -24 : days * 24)
		+ t->tm_hour - gmt.tm_hour);
	minutes=hours * 60 + t->tm_min - gmt.tm_min;
	return(minutes*60);
}
#endif

int cmpserver(const void *s1,const void *s2)
{
	return(strlen(((SERVER*)s2)->url)-strlen(((SERVER*)s1)->url));
}

/*
int FreeFilters(void){
int i;
	for(i=0; i < nfilters; i++){
		regfree(&(Filter[i].filter));
		free(Filter[i].regstr);
	}
	return(0);
}
*/

int AddFilter(char * filter,int filter_type,int reverse){
#define ERRSTRSIZE 100
char errstr[ERRSTRSIZE];
int err;
	if(nfilters<MAXFILTER){
		err=regcomp(&(Filter[nfilters].filter),filter,REG_EXTENDED|REG_ICASE);
		if(err){
			fprintf(stderr, "Config file error: wrong  regex: %s", filter);
			regerror(err, &(Filter[nfilters].filter), errstr, ERRSTRSIZE);
			fprintf(stderr, "Error is: %s", errstr);
			regfree(&(Filter[nfilters].filter));
			return(1);
		}
		Filter[nfilters].filter_type=filter_type;
		Filter[nfilters].reverse=reverse;
		Filter[nfilters].regstr=strdup(filter);
		nfilters++;
	}else{
		fprintf(stderr, "Too many allow/disallow/checkonly commands in config file\nTry to increment MAXFILTER in indexer.h and recompile");
		return(1);
	}
	return(0);
}

char FCBuf[64]="";
char *FilterComStr(UDM_FILTER* f){
	FCBuf[0]=0;
	if(!f)return(FCBuf);
	switch(f->filter_type){
		case F_ALLOW: sprintf(FCBuf,"Allow");break;
		case F_DISALLOW: sprintf(FCBuf,"Disallow");break;
		case F_HEAD: sprintf(FCBuf,"CheckOnly");break;
		default: sprintf(FCBuf,"Unknown");break;
	}
	if(f->reverse)strcat(FCBuf,"NoMatch");
	strcat(FCBuf," ");
	strcat(FCBuf,f->regstr?f->regstr:"");
	return(FCBuf);
}

/* Allow/Disallow/CheckOnly */
UDM_FILTER * WhatToDo(char *url){
int i;
#define NS 10
regmatch_t subs[NS];
int err;
	for(i=0;i<nfilters;i++){
		err=regexec(&(Filter[i].filter),url,NS,subs,0);
		if((Filter[i].reverse)&&(err))
			return(&Filter[i]);
		if((!Filter[i].reverse)&&(!err))
			return(&Filter[i]);
	}
	return(NULL);
}

int AddAlias(char * find,char *replace){
	if(naliases<MAXFILTER){
		Alias[naliases].find=strdup(find);
		Alias[naliases].replace=strdup(replace);
		naliases++;
	}else{
		fprintf(stderr, "Too many alias commands in config file\nTry to increment MAXFILTER in indexer.h and recompile");
		return(1);
	}
	return(0);
}


int AddServer(char *url,char *charset,char *proxy,
	int period,int hint,char *basic_auth,
	int outside,int maxhops,
	int gindex,int deletebad,int userobots,int gfollow,
	int bodyweight,int  titleweight,int  urlweight,
	int descweight,int  keywordweight,int max_net_errors,
	int read_timeout)
{
char *s;

	if(nservers>=mservers){
		if(mservers){
			mservers+=16;
			Server=(SERVER *)realloc(Server,mservers*sizeof(SERVER));
		}else{
			mservers=16;
			Server=(SERVER *)malloc(mservers*sizeof(SERVER));
		}
	}
	Server[nservers].url=strdup(url);
	Server[nservers].charset=charset?strdup(charset):NULL;
	Server[nservers].basic_auth=basic_auth?strdup(basic_auth):NULL;
	if(proxy){
		Server[nservers].proxy=strdup(proxy);
		s=strchr(Server[nservers].proxy,':');
		if(s){
			*s=0;
			Server[nservers].proxy_port=atoi(s+1)?atoi(s+1):DEFAULT_PROXY_PORT;
		}else{
			Server[nservers].proxy_port=DEFAULT_PROXY_PORT;
		}
	}else{
		Server[nservers].proxy=NULL;
		Server[nservers].proxy_port=0;
	}
	Server[nservers].period=period;
	Server[nservers].hint=hint;
	Server[nservers].net_errors=0;
	Server[nservers].outside=outside;
	Server[nservers].maxhops=maxhops;
	Server[nservers].gindex=gindex;
	Server[nservers].gfollow=gfollow;
	Server[nservers].deletebad=deletebad;
	Server[nservers].userobots=userobots;
	Server[nservers].bodyweight=bodyweight;
	Server[nservers].titleweight=titleweight;
	Server[nservers].urlweight=urlweight;
	Server[nservers].descweight=descweight;
	Server[nservers].keywordweight=keywordweight;
	Server[nservers].max_net_errors=max_net_errors;
	Server[nservers].read_timeout=read_timeout;
	nservers++;
	return(0);
}

int FindServer(char *url){
int i;
char *r;
	if(!Server)return(-1);
	url=remove2dot(url);
	r=strstr(url,"/robots.txt");
	/* If it's a robot.txt, cut to hostinfo and find result */
	if((r)&&(!strcmp(r,"/robots.txt"))){
		r++;*r=0;
		for(i=0;i<nservers;i++){
			if(!STRNCMP(url,Server[i].url)){
				*r='r';
				return(i);
			}
		}
		*r='r';
	}else{
		for(i=0;i<nservers;i++)
			if(!STRNCMP(url,Server[i].url))
				return(i);
	}
	return(-1);
}

int FindRobots(char *url){
ROBOT *r;
int i;
	if(Robots){
		r=Robots;i=0;
		while(r->hostinfo){
			if(strstr(url,r->hostinfo)&&strstr(url,r->path))
				return(i);
			i++;r++;
		}
	}
	return(-1);
}

int FreeRobots(){
ROBOT *r;
	if(!Robots)return(0);
	r=Robots;
	while(r->hostinfo){
		free(r->hostinfo);
		free(r->path);
		r++;
	}
	free_it(Robots);
	return(0);
}

int ParseRobots(char *content,char *hostinfo){
	int myrule=0;
	char *s,*e;

	DeleteRobotsFromHost(hostinfo);
	s=strtok(content,"\r\n");
	while(s){
		if(*s=='#'){
		}else
		if(!(STRNCASECMP(s,"User-Agent"))){
			if(strstr(s,INDEXER_USER_AGENT))myrule=1;
			else
			if(strstr(s,"*"))myrule=1;
			else
				myrule=0;
		}else
		if((!(STRNCASECMP(s,"Disallow")))&&(myrule)){
			if((e=strchr(s+9,'#')))*e=0;
			e=s+9;SKIP(e," \t");s=e;
			SKIPN(e," \t");*e=0;
			if(*s)AddRobotsToHost(hostinfo,s);
		}
		s=strtok(NULL,"\n\r");
	}
	return(0);
}


int AddWord(char *word,int where,int checkstop)
{
int i;
char *s;

	if(!word[0])return(0);
	s=strdup(word);
	tolower_string(s,lcindex);
	if(checkstop){
		if(IsStopWord(s)){
			free(s);return(0);
		}
	}
	for(i=0;i<wcur;i++){
		if(!strcmp(Word[i].word,s)){
			Word[i].count|=where;
			free(s);return(0);
		}
	}
	if(wcur>=wsize){
		if(wsize){
			wsize+=WSIZE;
			Word=(UDM_WORD *)realloc(Word,wsize*sizeof(UDM_WORD));
		}else{
			wsize=WSIZE;
			Word=(UDM_WORD *)malloc(wsize*sizeof(UDM_WORD));
		}
	}
	Word[wcur].word=s;
	Word[wcur].count=where;
	wcur++;return(0);
}

int FreeWords() {
	int i;
	for(i=0;i<wcur;i++) free(Word[i].word);
	wcur=0;
	return(0);
}

int AddHref(char *href,int referrer,int hops,int stored){
int i;
	if(!href[0])return(0);
	for(i=0;i<hcur;i++)
		if(!strcmp(Href[i].href,href)){
			Href[i].stored|=stored;
			return(0);
		}
	if(hcur>=hsize){
		if(hsize){
			hsize+=HSIZE;
			Href=(HREF *)realloc(Href,hsize*sizeof(HREF));
		}else{
			hsize=HSIZE;
			Href=(HREF *)malloc(hsize*sizeof(HREF));
		}
	}
	Href[hcur].href=strdup(href);
	Href[hcur].referrer=referrer;
	Href[hcur].hops=hops;
	Href[hcur].stored=stored;
	hcur++;return(0);
}

int FreeHrefs() {
int i;
	for(i=0;i<hcur;i++) free(Href[i].href);
	hcur=0;return(0);
}


int ShowStatistics(){
int expired_total=0;
int total_total=0;
STAT *stat,*s;

	printf("\n          UdmSearch statistics\n\n");
	printf("%10s %10s %10s\n","Status","Expired","Total");
	printf("   -----------------------------\n");

	if((stat=GetStatistics())){
		s=stat;
		while(stat->status!=-1){
			printf("%10d %10d %10d %s\n",
				stat->status,stat->expired,stat->total,
				HTTPErrMsg(stat->status));
			total_total+=stat->total;
			expired_total+=stat->expired;
			stat++;
		}
		free(s);
	}
	printf("   -----------------------------\n");
	printf("%10s %10d %10d\n\n","Total",expired_total,total_total);
	return(0);
}




int period=DEFAULT_REINDEX_TIME;
int maxhops=DEFAULT_MAX_HOPS;
int maxdocsize=MAXDOCSIZE;
int max_net_errors=MAXNETERRORS;
int read_timeout=READ_TIMEOUT;
int hint=0,gindex=1,gfollow=1,deletebad=1,userobots=1,outside=0;
int bodyweight=1,titleweight=2,keywordweight=2,descweight=2,urlweight=0;
int add_servers=1,force_charset_1251=0;
char *charset=NULL,*proxy=NULL,*basic_auth=NULL;
char user_agent[256]="";
char httpheaders[STRSIZ]="";


static int config_level=0;
int LoadConfig(char *config_name){
int i;
FILE *config;
URL from;
char *s,*s1, str[STRSIZ]="", str1[STRSIZ]="", str2[STRSIZ]="";

	if(!(config=fopen(config_name,"r"))){
		fprintf(stderr, "Error: can't open config file '%s': %s\r\n",
			config_name, strerror(errno));
		exit_nicely(1);
	}
	i=0;
	while(fgets(str,sizeof(str),config)){
		i++; strcpy(str1,"");
		if(!strcmp(str,"\n"))continue;
		if(str[0]=='#')continue;
		
		if(!STRNCASECMP(str,"MySQLHost")){
			if(sscanf(str+9,"%s",str1))
				DBHost=strdup(str1);
		}else
		if(!STRNCASECMP(str,"MySQLDB")){
			if(sscanf(str+7,"%s",str1)) 
				DBName=strdup(str1);
		}else
		if(!STRNCASECMP(str,"MySQLUser")){
			if(sscanf(str+9,"%s",str1)&&strcmp(str1,""))
				DBUser=strdup(str1);
		}else
		if(!STRNCASECMP(str,"MySQLPass")){
			if(sscanf(str+9,"%s",str1)&&strcmp(str1,""))
				DBPass=strdup(str1);
			else	DBPass=strdup("");
		}else
		if(!STRNCASECMP(str,"MySQLPort")){
			if(sscanf(str+9,"%d",&DBPort)!=1)DBPort=0;
		}else
		if(!STRNCASECMP(str,"DBHost")){
			if(sscanf(str+7,"%s",str1))
				DBHost=strdup(str1);
		}else
		if(!STRNCASECMP(str,"DBName")){
			if(sscanf(str+7,"%s",str1)) 
				DBName=strdup(str1);
		}else
		if(!STRNCASECMP(str,"DBUser")){
			if(sscanf(str+7,"%s",str1)&&strcmp(str1,""))
				DBUser=strdup(str1);
		}else
		if(!STRNCASECMP(str,"DBPass")){
			if(sscanf(str+7,"%s",str1)&&strcmp(str1,""))
				DBPass=strdup(str1);
			else	DBPass=strdup("");
		}else
		if(!STRNCASECMP(str,"DBPort")){
			if(sscanf(str+6,"%d",&DBPort)!=1)DBPort=0;
		}else
		if(!STRNCASECMP(str,"LocalCharset")){
			if(sscanf(str+13,"%s",str1)&&strcmp(str1,"")){
				free_it(localcharset);
				localcharset=strdup(str1);
				lcindex=getcharset(localcharset);
			}
		}else
		if(!STRNCASECMP(str,"DisallowNoMatch")){
			s=strtok(str," \t\r\n");
			while((s=strtok(NULL," \t\r\n"))){
				if(AddFilter(s,F_DISALLOW,1))
					exit_nicely(1);
			}
		}else
		if(!STRNCASECMP(str,"Disallow")){
			s=strtok(str," \t\r\n");
			while((s=strtok(NULL," \t\r\n"))){
				if(AddFilter(s,F_DISALLOW,0))
					exit_nicely(1);
			}
		}else
		if(!STRNCASECMP(str,"AllowNoMatch")){
			s=strtok(str," \t\r\n");
			while((s=strtok(NULL," \t\r\n"))){
				if(AddFilter(s,F_ALLOW,1))
					exit_nicely(1);
			}
		}else
		if(!STRNCASECMP(str,"Allow")){
			s=strtok(str," \t\r\n");
			while((s=strtok(NULL," \t\r\n"))){
				if(AddFilter(s,F_ALLOW,0))
					exit_nicely(1);
			}
		}else
		if(!STRNCASECMP(str,"CheckOnlyNoMatch")){
			s=strtok(str," \t\r\n");
			while((s=strtok(NULL," \t\r\n"))){
				if(AddFilter(s,F_HEAD,1))
					exit_nicely(1);
			}
		}else
		if(!STRNCASECMP(str,"CheckOnly")){
			s=strtok(str," \t\r\n");
			while((s=strtok(NULL," \t\r\n"))){
				if(AddFilter(s,F_HEAD,0))
					exit_nicely(1);
			}
		}else
		if(!STRNCASECMP(str,"AddType")){
			s=strtok(str," \t\r\n");
			if((s1=strtok(NULL," \t\r\n")))
				while((s=strtok(NULL," \t\r\n")))
					AddType(s1,s);
		}else
		if(!STRNCASECMP(str,"Tag")){
			if(sscanf(str+4,"%d",&hint)!=1)
				hint=0;
		}else
		if(!STRNCASECMP(str,"MaxNetErrors")){
			if(sscanf(str+12,"%d",&max_net_errors)!=1)
				max_net_errors=MAXNETERRORS;
		}else
		if(!STRNCASECMP(str,"ReadTimeOut")){
			if(sscanf(str+11,"%d",&read_timeout)!=1)
				read_timeout=READ_TIMEOUT;
		}else
		if(!STRNCASECMP(str,"Period")){
			if(sscanf(str+6,"%d",&period)!=1)
				period=DEFAULT_REINDEX_TIME;
		}else
		if(!STRNCASECMP(str,"CharSet")){
			free_it(charset);
			if(sscanf(str+7,"%s",str1))
				charset=strdup(str1);
		}else
		if(!STRNCASECMP(str,"Proxy")){
			free_it(proxy);
			if(sscanf(str+5,"%s",str1)&&strcmp(str1,""))
				proxy=strdup(str1);
		}else
		if(!STRNCASECMP(str,"HTTPHeader")){
			free_it(proxy);
			if(sscanf(str+11,"%[^\n\r]s",str1)&&strcmp(str1,"")){
				if(!STRNCMP(str1,"User-Agent: ")){
					strcpy(user_agent,str1+12);
				}else{
					strcat(httpheaders,str1);
					strcat(httpheaders,"\r\n");
				}
			}
		}else
#ifdef HAVE_SYSLOG_H 
		if(!STRNCASECMP(str,"SyslogFacility")){
			if(sscanf(str+14,"%s",str1)&&strcmp(str1,""))
				log_facility=syslog_facility(str1);
		}else
#endif
		if(!STRNCASECMP(str,"AuthBasic")){
			free_it(basic_auth);
			if(sscanf(str+9,"%s",str1)){
				basic_auth=(char*)malloc(BASE64_LEN(strlen(str1)));
				base64_encode(str1,basic_auth,strlen(str1));
			}
		}else
		if(!STRNCASECMP(str,"Mime")){
			AddMime(str+4);
		}else
		if(!STRNCASECMP(str,"Alias")){
			if(sscanf(str+5,"%s%s",str1,str2)==2)
				AddAlias(str1,str2);
			else{
				fprintf(stderr, "Error in config file '%s' line %d:%s",config_name,i,str);
				exit_nicely(1);
			}
		}else
		if(!STRNCASECMP(str,"Server")){
			if(sscanf(str+6,"%s",str1)==1){
				if(!parseurl(&from,str1)){
					AddServer(str1,charset,proxy,period,
					hint,basic_auth,outside,maxhops,
					gindex,deletebad,userobots,gfollow,
					bodyweight,titleweight,urlweight,
					descweight,keywordweight,
					max_net_errors,read_timeout);
					if((!strcmp(from.schema,"http"))&&(add_servers)){
						sprintf(str,"%s://%s/%s",
							from.schema,
							from.hostinfo,
							"robots.txt");
						AddHref(str,0,0,0);
					}
					if(add_servers)AddHref(str1,0,0,0);
				}else{
					fprintf(stderr,"File '%s' line %d: Invalid URL: %s",config_name,i,str1);
					exit_nicely(1);
				}
			}
		}else
		if(!STRNCASECMP(str,"FollowOutside")){
			if(sscanf(str+13,"%s",str1)==1){
				if(!STRNCASECMP(str1,"yes"))
					outside=1;
				else
					outside=0;
			}else
				outside=0;
		}else
		if(!STRNCASECMP(str,"Index")){
			if(sscanf(str+5,"%s",str1)==1){
				if(!STRNCASECMP(str1,"yes"))
					gindex=1;
				else
					gindex=0;
			}else
				gindex=1;
		}else
		if(!STRNCASECMP(str,"Follow ")){
			if(sscanf(str+7,"%s",str1)==1){
				if(!STRNCASECMP(str1,"yes"))
					gfollow=1;
				else
					gfollow=0;
			}else
				gfollow=1;
		}else
		if(!STRNCASECMP(str,"Robots")){
			if(sscanf(str+6,"%s",str1)==1){
				if(!STRNCASECMP(str1,"yes"))
					userobots=1;
				else
					userobots=0;
			}else
				userobots=1;
		}else
		if(!STRNCASECMP(str,"DeleteBad")){
			if(sscanf(str+9,"%s",str1)==1){
				if(!STRNCASECMP(str1,"yes"))
					deletebad=1;
				else
					deletebad=0;
			}else
				deletebad=1;
		}else
		if(!STRNCASECMP(str,"MaxHops")){
			if(sscanf(str+7,"%d",&maxhops)!=1)
				maxhops=0;
		}else
		if(!STRNCASECMP(str,"BodyWeight")){
			if(sscanf(str+10,"%d",&bodyweight)!=1)
				bodyweight=1;
		}else
		if(!STRNCASECMP(str,"TitleWeight")){
			if(sscanf(str+11,"%d",&titleweight)!=1)
				titleweight=2;
		}else
		if(!STRNCASECMP(str,"DescWeight")){
			if(sscanf(str+10,"%d",&descweight)!=1)
				descweight=2;
		}else
		if(!STRNCASECMP(str,"UrlWeight")){
			if(sscanf(str+9,"%d",&urlweight)!=1)
				urlweight=0;
		}else
		if(!STRNCASECMP(str,"KeywordWeight")){
			if(sscanf(str+13,"%d",&keywordweight)!=1)
				keywordweight=2;
		}else
		if(!STRNCASECMP(str,"MaxDocSize")){
			if(sscanf(str+11,"%d",&maxdocsize)!=1)
				maxdocsize=MAXDOCSIZE;
		}else
		if(!STRNCASECMP(str,"NoIndex")){
			gindex=0;
		}else
		if(!STRNCASECMP(str,"NoFollow")){
			gfollow=0;
                }else
                if(!STRNCASECMP(str,"ForceIISCharset1251")){
			if(sscanf(str+20,"%s",str1)==1){
				if(!STRNCASECMP(str1,"yes"))force_charset_1251=1;
				else	force_charset_1251=0;
			}else	force_charset_1251=0;
		}else
		if(!STRNCASECMP(str,"Include")){
			if((sscanf(str+7,"%s",str1))&&(config_level<5)){
				if(*str1=='/')strcpy(str,str1);
				else	sprintf(str,"%s/%s",CONFIGDIR,str1);
				LoadConfig(str);
			}
		}else{
			fprintf(stderr,"Error in config file '' line %d:%s\n",i,str);
			exit_nicely(1);
		}
	}
	fclose(config);return(0);
}


int usage(){
	fprintf(stderr,
"\n\
indexer from UdmSearch v." VERSION "/%s\n\
http://mysearch.udm.net (C) 1998-1999, UdmSearch Developers Team.\n\
\n\
Usage: indexer [OPTIONS]  [configfile]\n\
\n\
Indexing options:\n\
  -a            reindex All documents (even if not expired)\n\
  -m            reindex if not Modified\n\
  -e            more Expired first\n\
  -n number     index only Number documents and exit\n\
  -p number     sleeP number seconds after each URL\n\
  -q            quick startup (do not add Server URLs)\n\
\n\
Subsection control options:\n\
  -t tag        set filter on Tag\n\
  -u pattern    set filter on URL (SQL LIKE wildcards)\n\
  -s status     set filter on HTTP Status\n\
\n\
Logging options:\n\
"
#ifdef LOG_PERROR
"  -l            do not Log to stdout/stderr\n\
"
#endif
"  -v number     Verbose level, 0-5\n\
\n\
Ispell import options:\n\
  -L language   Two letters Language code (en, ru, de etc.)\n\
  -A filename   ispell Affix file\n\
  -D filename   ispell Dictionary file\n\
  -d            dump to stdout instead of storing to database\n\
\n\
Mics. options:\n\
  -C            Clear database and exit\n\
  -S            print Statistics and exit\n\
  -I            print referers and exit\n\
  -h,-?         print this Help page and exit\n\
\n\
\n\
Mail bug reports and suggestions to <devel@search.udm.net>.\n\
",DBVersion);
	return(0);
}

int ParseText(char *content,char *Sep,int weight,int check){
char *s;
	s=strtok(content,Sep);
	while(s){
		if(weight)
			AddWord(s,weight,check);
		s=strtok(NULL,Sep);
	}return(0);
}

int main(int argc, char **argv) {
extern char *optarg;
extern int optind;
char str[STRSIZ]="",str1[STRSIZ]="";
char text[2*MAXTEXTSIZE]="";      char text_escaped[2*MAXTEXTSIZE+1]="";
char keywords[MAXKEYWORDSIZE]=""; char keywords_escaped[2*MAXKEYWORDSIZE+1]="";
char descript[MAXDESCSIZE]="";    char descript_escaped[2*MAXDESCSIZE+1]="";
char last_modified[32]="";        char last_modified_escaped[32*2+1]="";
char title[MAXTITLESIZE]="";      char title_escaped[MAXTITLESIZE*2+1];
char subj[MAXTITLESIZE]="";
char from[MAXKEYWORDSIZE]="";
char wordchar[256]="";		  char digest[33];
URL curURL,newURL,realURL;
TAG tag;DOCUMENT * doc;

char *buf,*s,*e,*tmp,sch;
char *cururl,*last_index_time;
char *content,*header,*content_type,*location,*href,*statusline,*crc;
char *mime,*newschema;

char *language=NULL,*affix=NULL,*dictionary=NULL;
char *config_name= CONFIGDIR "/indexer.conf";

int inbody,intitle,inscript,follow,index,changed,ch,i,j,state,len,mimeno;
int npages=-1,npages0=0,expired_first=0;
int reindexall=0,clear=0,stat=0,integrity=0,reload_robots=1;
int gmt_offset=0,origin=0,clone=0;
int url_id,size,oldsize,oldstatus,status,hops,realsize;
int Method=0,dump=0,seconds=0;
struct tm *tim;time_t clock;
int log2stderr=1, reindex=0;
UDM_FILTER *MatchedFilter=NULL;

	init_charset();

#ifdef HAVE_TM_GMTOFF
	clock=time(0);tim=localtime(&clock);
	gmt_offset=tim->tm_gmtoff;
#else
	gmt_offset=get_gmtoff();
#endif

	/* Parse arguments */
	while ((ch = getopt(argc, argv, "CSIahelmdq?t:u:s:n:v:L:A:D:p:")) != -1){
		switch (ch) {
		case 'C': clear++;add_servers=0;break;	
		case 'S': stat++;add_servers=0;break;
		case 'I': integrity++;add_servers=0;break;
		case 'L': language=optarg;break;
		case 'A': affix=optarg;add_servers=0;break;
		case 'D': dictionary=optarg;add_servers=0;break;
		case 'q': add_servers=0;break;
		case 'l': log2stderr=0;break;
		case 'a': reindexall=1;break;
		case 'e': expired_first=1;break;
		case 'm': reindex=1;break;
		case 'n': npages=atoi(optarg);npages0=npages;break;
		case 'v': loglevel=atoi(optarg);break;
		case 'p': seconds=atoi(optarg);break;
		case 'd': dump=1;break;
		case 't': AddTagLimit(atoi(optarg));break;
		case 's': AddStatusLimit(atoi(optarg));break;
		case 'u': AddURLLimit(optarg);break;
		case '?':
		case 'h':
		default:  exit_nicely(usage());
		}
	}
	argc -= optind;
	argv += optind;
	if(argc>1){
		usage();
		exit_nicely(1);
	}
	if (argc==1)config_name=argv[0];
	sprintf(user_agent,"%s/%s",INDEXER_USER_AGENT,VERSION);

	LoadConfig(config_name);

	makewordchar(wordchar,lcindex);
	s=SEPARATOR;*s=0;
	for(i=1;i<255;i++){
		if(!strchr(wordchar,i)){
			*s=i;s++;*s=0;
		}
	}
	if(affix||dictionary){
		if(!language){
			fprintf(stderr,"Language is not specified for import!\n");
			exit_nicely(1);
		}
		if(strlen(language)!=2){
			fprintf(stderr,"Language should be 2 letters\n");
			exit_nicely(1);
		}
		if(affix)
			exit_nicely(ImportAffixes(language,affix,dump));
		if(dictionary)
			exit_nicely(ImportDictionary(language,dictionary,dump));
	}
	if(clear)exit_nicely(ClearDatabase());
	if(stat)exit_nicely(ShowStatistics());
	if(integrity)exit_nicely(ShowReferers());
	if(reindexall)MarkForReindex();

	/* Add one virtual server if we want to follow outside */
	if(outside)
	AddServer("",charset,proxy,period,hint,
		basic_auth,outside,maxhops,gindex,
		deletebad,userobots,gfollow,
		bodyweight,titleweight,urlweight,
		descweight,keywordweight,
		max_net_errors,read_timeout);

	free_it(charset);free_it(proxy);free_it(basic_auth);

/***************************************************************/
/****************** Now start indexing ... *********************/
/***************************************************************/

	/*  Long name should be found first    */
	/*  to allow different options         */
	/*  for server and it's subdirectories */
	qsort((void*)Server,nservers,sizeof(SERVER),cmpserver);

	/* Save URLs from indexer.conf */
	StoreHrefs(hcur,Href);
	if(hcur>MAXHSIZE)FreeHrefs();
	LoadStopList();

	/* Alloc buffer for document */
	buf=(char*)malloc(maxdocsize);

	/* Open log */
	udm_openlog(log_facility,log2stderr);
	udm_log(INFO, "indexer from UdmSearch v." VERSION"/%s started with config %s", DBVersion,config_name);

while(1){
	if(!npages){
		udm_log(INFO,"%d documents done",npages0);
		exit_nicely(0);
	}
	if(npages>0)npages--;

	if(reload_robots){
		FreeRobots();
		Robots=LoadRobots(Robots);
		reload_robots=0;
	}

	status=1;changed=1;
	inbody=inscript=intitle=clone=0;
	content_type=location=href=charset=statusline=NULL;
	cururl=0;last_index_time=0;digest[0]=0;
	text[0]=0;title[0]=0;keywords[0]=0;descript[0]=0;subj[0]=0;
	last_modified[0]=0;from[0]=0;

	doc=GetDocInfo(gmt_offset,expired_first);

	if(!doc){
		StoreHrefs(hcur,Href);
		doc=GetDocInfo(gmt_offset,expired_first);
	}
	if(!doc){
		udm_log(INFO,"Done");
		exit_nicely(0);
	}
	cururl=doc->url;
	url_id=doc->url_id;
	oldsize=doc->size;
	oldstatus=doc->status;
	last_index_time=doc->last_index_time;
	hops=doc->hops;
	crc=doc->crc;
	strcpy(last_modified,doc->last_modified?doc->last_modified:"");

	AddHref(cururl,0,hops,1);
	udm_log(INFO,"%s",cururl);

#ifdef HAVE_SETPROCTITLE
	setproctitle("%s",cururl);
#endif
	if(parseurl(&curURL,cururl)){
		DeleteUrl(url_id);
		udm_log(WARNING,"Invalid URL: %s ... deleted",cururl);
		freeurl(&curURL);
		continue;
	}
	if((i=FindServer(cururl))<0){
		DeleteUrl(url_id);
		if(!strcmp(curURL.filename,"robots.txt")){
			DeleteRobotsFromHost(curURL.hostinfo);
			reload_robots=1;
		}
		udm_log(WARNING,"No 'Server' command for url... deleted.");
		freeurl(&curURL);
		continue;
	}
	period=Server[i].period;
	hint=Server[i].hint;
	outside=Server[i].outside;
	maxhops=Server[i].maxhops;
	gindex=Server[i].gindex;index=gindex;
	gfollow=Server[i].gfollow;follow=gfollow;
	userobots=Server[i].userobots;
	deletebad=Server[i].deletebad;
	bodyweight=Server[i].bodyweight;
	titleweight=Server[i].titleweight;
	urlweight=Server[i].urlweight;
	descweight=Server[i].descweight;
	keywordweight=Server[i].keywordweight;
	read_timeout=Server[i].read_timeout;

	charset=Server[i].charset;
	if(hops>maxhops){
		DeleteUrl(url_id);
		udm_log(WARNING,"Too many hops (%d)... deleted.",hops);
		freeurl(&curURL);
		continue;
	}
	if(!(MatchedFilter=WhatToDo(cururl)))
		Method=F_ALLOW;
	else
		Method=MatchedFilter->filter_type;

	if(Method==F_DISALLOW){
		DeleteUrl(url_id);
		if(!strcmp(curURL.filename,"robots.txt")){
			DeleteRobotsFromHost(curURL.hostinfo);
			reload_robots=1;
		}
		udm_log(WARNING,"URL deleted : '%s'",FilterComStr(MatchedFilter));
		freeurl(&curURL);
		continue;
	}
	if((userobots)&&(FindRobots(cururl)>=0)){
		/* if robots.txt disallowed by himself */
		/* we will not delete it */
		if(strcmp(curURL.filename,"robots.txt")){
			DeleteUrl(url_id);
			udm_log(WARNING,"URL deleted: robots.txt");
		}
		freeurl(&curURL);
		continue;
	}
	
	if((!strcmp(curURL.schema,"ftp"))&&(!Server[i].proxy)){
		udm_log(WARNING,"FTP is available via proxy only, skipped");
		UpdateUrl(url_id,504,period);
		freeurl(&curURL);continue;
	}
	if(Server[i].net_errors>Server[i].max_net_errors){
		udm_log(WARNING,"Too many network errors for this server, skipped");
		UpdateUrl(url_id,504,period);
		freeurl(&curURL);continue;
	}

	/* Find alias */
	parseurl(&realURL,cururl);
	for(j=0;j<naliases;j++){
		if(!STRNCMP(cururl,Alias[j].find)){
			sprintf(str1,"%s%s",Alias[j].replace,cururl+strlen(Alias[j].find));
			freeurl(&realURL);
			if(parseurl(&realURL,str1)){
				udm_log(UDM_LOG_ERROR,"Error in aliased URL: '%s'",str1);
				freeurl(&curURL);continue;
			}
			break;
		}
	}


	/* Compose HTTP header */
	sprintf(str,"%s %s%s HTTP/1.0\r\n",
		(Method==F_HEAD)?"HEAD":"GET",
		Server[i].proxy?cururl:realURL.path,
		Server[i].proxy?"":realURL.filename);

#if (HAVE_MSQL||HAVE_IODBC||HAVE_UNIXODBC||HAVE_SOLID)
	if((oldstatus)&&(last_modified[0])&&(!reindex))
		sprintf(STREND(str),"If-Modified-Since: %s\r\n",
			last_modified);
#else  /* PgSQL MySQL */
	if((oldstatus)&&(last_index_time[0])&&(!reindex))
		sprintf(STREND(str),"If-Modified-Since: %s\r\n",
			last_index_time);
#endif

	if((!strcmp(realURL.schema,"http"))||(!strcmp(realURL.schema,"ftp"))){

		/* User agent */
		sprintf(STREND(str),"User-Agent: %s\r\n",user_agent);
		
		/* If LocalCharset specified */
		if(localcharset)
			sprintf(STREND(str),"Accept-charset: %s\r\n",
				localcharset);

		sprintf(STREND(str),"Host: %s\r\n",realURL.hostname);

		/* Auth if required */
		if(Server[i].basic_auth){
			strcat(str,"Authorization: Basic ");
			strcat(str,Server[i].basic_auth);
			strcat(str,"\r\n");
		}
	}

	/* Add user defined headers */
	if(*httpheaders)strcat(str,httpheaders);

	/* Empty line is the end of HTTP header */
	strcat(str,"\r\n");
	
	
	if(!strcmp(realURL.schema,"file")){
		size = FILEGet(str,buf,maxdocsize);
	}else
	if((!strcmp(realURL.schema,"news"))){
		size = NEWSGet(realURL.hostname,realURL.port?realURL.port:119,
			str,buf,maxdocsize,read_timeout
		);
	}else
	if((!strcmp(realURL.schema,"nntp"))){
		size = NNTPGet(realURL.hostname,realURL.port?realURL.port:119,
			str,buf,maxdocsize,read_timeout
		);
	}else
	if((!strcmp(realURL.schema,"http"))){
		size = HTTPGet(
			Server[i].proxy?Server[i].proxy:realURL.hostname,
			Server[i].proxy?Server[i].proxy_port:(realURL.port?realURL.port:80),
			str,buf,maxdocsize,read_timeout
		);
	}else
	if((!strcmp(realURL.schema,"ftp"))){
		size = HTTPGet(
			Server[i].proxy?Server[i].proxy:realURL.hostname,
			Server[i].proxy?Server[i].proxy_port:(realURL.port?realURL.port:80),
			str,buf,maxdocsize,read_timeout
		);
	}else{
		DeleteUrl(url_id);
		udm_log(WARNING,"Unknown URL schema: %s",cururl);
		freeurl(&curURL);continue;
	}

	/* Timeout */
	if(size==-3){
		udm_log(WARNING,"Download timeout");
		UpdateUrl(url_id,504,period);
		Server[i].net_errors++;
		freeurl(&curURL);freeurl(&realURL);continue;
	}

	/* No connection */
	if(size<=0){
		udm_log(WARNING,"Could not connect to host %s:%d",
			Server[i].proxy?Server[i].proxy:curURL.hostname,
			Server[i].proxy?Server[i].proxy_port:(curURL.port?curURL.port:80));
		UpdateUrl(url_id,503,period);
		Server[i].net_errors++;
		freeurl(&curURL);freeurl(&realURL);continue;
	}
	buf[size]=0;header=buf;realsize=size;

	/* Parse response header */
	if(!(content=strstr(buf,"\r\n\r\n")))content=strstr(buf,"\n\n");

	if(!content){
		udm_log(UDM_LOG_ERROR,"URL has no content, skipped");
		UpdateUrl(url_id,503,period);
		Server[i].net_errors++;
		freeurl(&curURL);freeurl(&realURL);continue;
	}
	*content=0; content+=4; /* !!! fixme */
	size-=(content-buf);    /* Could be changes */
	realsize-=(content-buf); /* Will be safe for DIGEST, Parsers etc */


	s=strtok(header,"\r\n");
	while(s){
		if(!STRNCASECMP(s,"HTTP/")){
			status=atoi(s+8);statusline=s;
		}else
		if (force_charset_1251 && !STRNCASECMP(s,"Server: ")){
                	if (sscanf(s+7,"%s",str1)==1)
			    if ( !STRNCASECMP(str1,"Microsoft") || !STRNCASECMP(str1,"IIS") )
               	              charset="windows-1251";
		}else		
		if(!STRNCASECMP(s,"Content-Type: ")){
			content_type=s+14;
			if((charset=strstr(content_type,"charset=")))
				charset+=8;
		}else
		if(!STRNCASECMP(s,"Location: ")){
			location=s+10;
		}else
		if(!STRNCASECMP(s,"Content-Length: ")){
			size=atoi(s+16);
		}else
		if(!STRNCASECMP(s,"Subject: ")){
			STRNCPY(title,s+9);strcpy(subj,title);
		}else
		if(!STRNCASECMP(s,"From: ")){
			STRNCPY(from,s+6);
		}else
		if(!STRNCASECMP(s,"Newsgroups: ")){
			STRNCPY(keywords,s+12);
		}else
		if(!STRNCASECMP(s,"Last-Modified: ")){
			STRNCPY(last_modified,s+15);
			escstr(last_modified_escaped,last_modified);
		}
		udm_log(DEBUG,"%s",s);
		s=strtok(NULL,"\r\n");
	}
	udm_log(EXTRA,"%s %s %d",statusline?statusline:"?",content_type?content_type:"?",size);


	if(status==1){ /* No HTTP code */
		UpdateUrl(url_id,status,period);
		udm_log(UDM_LOG_ERROR,"No http code!");
		Server[i].net_errors++;
		freeurl(&curURL);freeurl(&realURL);continue;
	}

	switch(HTTPResponseType(status)){
	case HTTP_STATUS_OK:
		if(!content_type){
			UpdateUrl(url_id,status,period);
			udm_log(UDM_LOG_ERROR,"No Content-type in '%s'!",cururl);
			Server[i].net_errors++;
			freeurl(&curURL);freeurl(&realURL);continue;
		}
		break;

	case HTTP_STATUS_REDIRECT: /* We'll try to use Location: xxx instead */
		DeleteWordFromURL(url_id);

		if((hops<maxhops)&&(location)){
			if((MatchedFilter=WhatToDo(location)))
				Method=MatchedFilter->filter_type;
			else
				Method=F_ALLOW;
			if(Method!=F_DISALLOW){
				if(!parseurl(&newURL,location)){
					if(FindServer(location)>=0)
						AddHref(location,url_id,hops+1,0);
					freeurl(&newURL);
				}else{
					udm_log(UDM_LOG_ERROR,"Error in redirect URL: '%s'",location);
				}
			}
		}
		UpdateUrl(url_id,status,period);
		freeurl(&curURL);freeurl(&realURL);continue;

	case HTTP_STATUS_NOT_MODIFIED:  /* Not Modified, nothing to do */
		UpdateUrl(url_id,status,period);
		freeurl(&curURL);freeurl(&realURL);continue;

	case HTTP_STATUS_DELETE:		
		if(deletebad){
			/* delete it if not robots.txt */
			if(!strcmp(curURL.filename,"robots.txt")){
				UpdateUrl(url_id,status,period);
				DeleteRobotsFromHost(curURL.hostinfo);
				reload_robots=1;
			}else{
				udm_log(EXTRA,"URL deleted");
				DeleteUrl(url_id);
			}
		}else{
			UpdateUrl(url_id,status,period);
		}
		freeurl(&curURL);freeurl(&realURL);continue;

	case HTTP_STATUS_RETRY: /* We'll retry later, maybe host is down */
		udm_log(EXTRA,"Could not read URL, skipped");
		UpdateUrl(url_id,status,period);
		Server[i].net_errors++;
		freeurl(&curURL);freeurl(&realURL);continue;

	default: /* Unknown status, retry later */
		udm_log(WARNING,"HTTP %d We don't yet know how to handle it, skipped",status);
		UpdateUrl(url_id,status,period);
		freeurl(&curURL);freeurl(&realURL);continue;
	}

	/* Now we have HTTP_OK and know Content-Type */

	/* Let's try to start external parser for this Content-Type */
	if((mime=mime_parse(content_type,&mimeno,content,realsize,maxdocsize-(content-buf)))){
		content=mime;content_type=mime_name(mimeno,MIME_TO);
		if((mime=mime_charset(mimeno)))charset=mime;
	};

	/* robots.txt */
	if(!STRNCASECMP(content_type,"text/plain")&&
	(!strcmp(curURL.filename,"robots.txt"))){
		ParseRobots(content,curURL.hostinfo);
		reload_robots=1;
	}else

	/* plain TEXT or the same */
	if(!STRNCASECMP(content_type,"text/plain")||
	!STRNCASECMP(content_type,"text/tab-separated-values")||
	!STRNCASECMP(content_type,"text/css"))
	{
		recode_string(content,getcharset(charset),lcindex);
		MD5_DIGEST(content,realsize,digest);
		changed=strcmp(digest,crc);
		if((origin=FindOrigin(digest))&&(origin!=url_id))clone=1;

		if(((index)&&(!clone)&&(changed))||reindex){
			strncpy(text,content,sizeof(text));
			text[sizeof(text)-1]=0;
			ParseText(content,SEPARATOR,bodyweight,1);
		}
	}else

	/* HTML text */
	if(!STRNCASECMP(content_type,"text/html")){

		recode_string(content,getcharset(charset),lcindex);
		MD5_DIGEST(content,realsize,digest);
		changed=strcmp(digest,crc);
		if((origin=FindOrigin(digest))&&(origin!=url_id))clone=1;
		
		state=HTML_UNK; s=content;
		while((*s)&&(changed||reindex)){
		switch(state){
		case HTML_UNK: /* yet unknown */
			if(!STRNCMP(s,"<!--"))
				state=HTML_COM;
			else
			if(*s=='<')
				state=HTML_TAG;
			else	
				state=HTML_TXT;
			break;

		case HTML_COM: /* comment */
			e=strstr(s,"-->");
			if(e)len=e-s+3;
			else len=strlen(s);
			s=s+len; state=HTML_UNK;
			break;

		case HTML_TAG: /* tag */
			e=strchr(s,'>');
			if(e)len=e-s+1;
			else len=strlen(s);
			tmp=(char*)malloc(len+1);
			strncpy(tmp,s,len);tmp[len]=0;
			parsetag(&tag,tmp);

			if(!strcasecmp(tag.name,"body"))inbody=1;
			else
			if(!strcasecmp(tag.name,"/body"))inbody=0;
			else
			if(!strcasecmp(tag.name,"script"))inscript=1;
			else
			if(!strcasecmp(tag.name,"/script"))inscript=0;
			else
			if(!strcasecmp(tag.name,"title"))intitle=1;
			else
			if(!strcasecmp(tag.name,"/title"))intitle=0;
			else
			if((!strcasecmp(tag.name,"a"))&&(tag.href))
				href=strdup(tag.href);
			else
			if((!strcasecmp(tag.name,"area"))&&(tag.href))
				href=strdup(tag.href);
			else
			if((!strcasecmp(tag.name,"frame"))&&(tag.src))
				href=strdup(tag.src);
			else
			if((!strcasecmp(tag.name,"meta"))&&(tag.equiv)&&(tag.content)){
				if((!strcasecmp(tag.equiv,"Content-Type"))&&(!charset)){
					if((charset=strstr(tag.content,"charset="))){
						charset+=8;
						recode_string(content,getcharset(charset),lcindex);
						/*
						Is it required here?
						MD5_DIGEST(content,realsize,digest);
						changed=strcmp(digest,crc);
						if((origin=FindOrigin(digest))&&(origin!=url_id))clone=1;
						*/
					}
				}else
				if(!strcasecmp(tag.equiv,"keywords")){
					strncpy(keywords,tag.content,sizeof(keywords));
					keywords[sizeof(keywords)-1]=0;
					if(index&&keywordweight){
						strcpy(str,keywords);
						ParseText(str,SEPARATOR,keywordweight,1);
					}
				}else
				if(!strcasecmp(tag.equiv,"description")){
					strncpy(descript,tag.content,sizeof(descript));
					descript[sizeof(descript)-1]=0;
					if(index&&descweight)
						strcpy(str,descript);
						ParseText(str,SEPARATOR,descweight,1);
				}else
				if(!strcasecmp(tag.equiv,"robots")){
					if(userobots){
					e=strtok(tag.content,SEPARATOR);
					while(e){
						if(!strcasecmp(e,"ALL")){
							follow=1;index=1;
						}else
						if(!strcasecmp(e,"NONE")){
							follow=0;index=0;
						}else
						if(!strcasecmp(e,"NOINDEX"))
							index=0;
						else
						if(!strcasecmp(e,"NOFOLLOW"))
							follow=0;
						else
						if(!strcasecmp(e,"INDEX"))
							index=1;
						else
						if(!strcasecmp(e,"FOLLOW"))
							follow=1;
						e=strtok(NULL,SEPARATOR);
					}}
				}
			}
			if((href)&&(hops<maxhops)&&(follow)&&(gfollow)){
				if(!parseurl(&newURL,href)){
					if(newURL.schema[0])newschema=newURL.schema;
					else	newschema=curURL.schema;
					if(!strcmp(newschema,"file")){
						sprintf(str,"file:%s%s",newURL.path[0]?newURL.path:curURL.path,newURL.filename);
					}else{
						sprintf(str,"%s://%s%s%s",
						newURL.schema[0]?newURL.schema:curURL.schema,
						newURL.hostinfo[0]?newURL.hostinfo:curURL.hostinfo,
						newURL.path[0]?newURL.path:curURL.path,newURL.filename);
					}
					remove2dot(str);
					if(!STRNCMP(str,"ftp://")&&(strstr(str,";type=")))
						*(strstr(str,";type"))=0;
		
					if((MatchedFilter=WhatToDo(str)))
						Method=MatchedFilter->filter_type;
					else
						Method=F_ALLOW;

					if((Method!=F_DISALLOW)
						&&((FindRobots(str)<0)||(!userobots))
						&&(FindServer(str)>=0)){
						if((!strcmp(newURL.schema,"http"))&&(outside)){
							sprintf(str1,"%s://%s/%s",newURL.schema,newURL.hostinfo,"robots.txt");
							AddHref(str1,0,0,0);
						}
						AddHref(str,url_id,hops+1,0);
					}else{
						udm_log(EXTRA,"URL '%s' not added: '%s'",str,FilterComStr(MatchedFilter));
					}
					freeurl(&newURL);
				}
				free_it(href);
			}
			freetag(&tag);free(tmp);
			s=s+len; state=HTML_UNK;
			break;

		case HTML_TXT: /* text */
			e=strchr(s,'<');
			if(e) len=e-s;
			else  len=strlen(s);
			tmp=(char*)malloc(2*len+6);
			e=tmp;j=0;*e=0;
			for(i=0;i<len;i++){
				switch(s[i]){
				case ' ' : case '\t':
				case '\n': case '\r':
					j++; break;

				case '&': /* parse specials */
					if(j){*e=' ';e++;j=0;}
					i++;
					if((sch=SgmlToChar(s+i,lcindex))){
						if(sch==' ')j++;
						*e=sch;e++;*e=0;
						while((i<len)&&(isalnum(*(s+i+1))))
							i++;
						if(s[i+1]==';')i++;
					}else{
						*e='?';e++;*e=0;
					}
					break;
				default:
					if(j){*e=' ';e++;j=0;}
					*e=(s[i]=='\'')?'`':s[i];
					e++; *e=0;
					break;
				}
			}

			if((inbody&&!inscript)&&(strlen(text)<MAXTEXTSIZE)){
				if(*text)strcat(text," ");
				strncat(text,tmp,MAXTEXTSIZE);
			}

			if((intitle)&&(strlen(title)<MAXTITLESIZE)){
				if(*title)strcat(title," ");
				strncat(title,tmp,MAXTITLESIZE);
				title[sizeof(title)-1]=0;
			}

			if(index&&!inscript){
				ParseText(tmp,SEPARATOR,
					bodyweight*inbody+titleweight*intitle,1);
			}
			free(tmp);
			s=s+len; state=HTML_UNK;
			break;

		}} /* switch and while */

	}else{
		index=0;
		MD5_DIGEST(content,realsize,digest);
		changed=strcmp(digest,crc);
		if((origin=FindOrigin(digest))&&(origin!=url_id))clone=1;
	}

	if((urlweight)&&(strcmp(curURL.filename,"robots.txt")))
		ParseText(cururl,SEPARATOR,urlweight,1);

	if(clone){
		DeleteWordFromURL(url_id);
		udm_log(EXTRA,"Duplicate document");
		MarkClone(url_id,status,period,content_type,last_modified_escaped,digest);
		freeurl(&curURL);freeurl(&realURL);
		FreeWords();continue;
	}else{
		if((!changed)&&(!reindex)){
			UpdateUrl(url_id,status,period);
			freeurl(&curURL);freeurl(&realURL);
			FreeWords();continue;
		}
		ParseText(subj,SEPARATOR,titleweight,1);
	}

	StoreHrefs(hcur,Href);if(hcur>MAXHSIZE)FreeHrefs();
	StoreWords(url_id,wcur,Word);FreeWords();

	/* Prepare text,keywords,description */
	if(!strcmp(realURL.schema,"news"))strcpy(keywords,from);

	s=text;while(*s){if(strchr("\t\n\r",*s))*s=' ';s++;}
	s=keywords;while(*s){if(strchr("\t\n\r",*s))*s=' ';s++;}
	s=descript;while(*s){if(strchr("\t\n\r",*s))*s=' ';s++;}
	s=title;while(*s){if(strchr("\t\n\r",*s))*s=' ';s++;}
	escstr(text_escaped,text);
	escstr(keywords_escaped,keywords);
	escstr(descript_escaped,descript);
	escstr(title_escaped,title);

#if (HAVE_MSQL||HAVE_IODBC)
	text_escaped[MAXTEXTSIZE]=0;
#endif

	LongUpdateUrl(url_id,status,changed,size,period,hint,index,
		last_modified_escaped,
		text_escaped,
		title_escaped,
		content_type,
		keywords_escaped,
		descript_escaped,
		digest);

	freeurl(&curURL);freeurl(&realURL);
	if(seconds){
		udm_log(DEBUG,"Sleeping %d second(s)",seconds);
		sleep(seconds);
	}

} /* endwhile */
} /* end of program */
