#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <netdb.h>
#include <signal.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <dirent.h>
#include <errno.h>
#include <string.h>
#include <regex.h>

#include "proto.h"
#include "common.h"

#define MAXMIME 128
typedef struct mime_struct {
	regex_t	reg;
	char	*mime_type;
} MIME;
static int nmimes=0;
static MIME Mime[MAXMIME];


int AddType(char * mime_type,char *reg){
#define ERRSTRSIZE 100
char errstr[ERRSTRSIZE];
int err;
	if(nmimes<MAXMIME){
		Mime[nmimes].mime_type=strdup(mime_type);
		err=regcomp(&(Mime[nmimes].reg),reg,REG_EXTENDED|REG_ICASE);
		if(err){
			fprintf(stderr, "Config file error: wrong  regex: %s", reg);
			regerror(err, &(Mime[nmimes].reg), errstr, ERRSTRSIZE);
			fprintf(stderr, "Error is: %s", errstr);
			regfree(&(Mime[nmimes].reg));
			return(1);
		}
		nmimes++;
	}else{
		fprintf(stderr, "Too many AddType\n");
		return(1);
	}
	return(0);
}

static char * ContentType(char *url){
int i;
#define NS 10
regmatch_t subs[NS];
int err;
	for(i=0;i<nmimes;i++){
		err=regexec(&(Mime[i].reg),url,NS,subs,0);
		if(!err)return(Mime[i].mime_type);
	}
	return(NULL);
}


static int alarm_abort;
static void sig_alarm(int i){ 
	alarm_abort = 1;
}

/* Open TCP Connect */
int open_host(char *hostname,int port, int timeout)
{
int net;
struct hostent *host;
struct sockaddr_in sin;
void *old_alarm;


 bzero( (char*)&sin, sizeof(sin) );
 if ((sin.sin_addr.s_addr=inet_addr(hostname)) != (unsigned long) -1)
   sin.sin_family=AF_INET;
 else
   if ((host=gethostbyname(hostname))) {
     sin.sin_family=host->h_addrtype;
     memcpy((caddr_t)&sin.sin_addr, host->h_addr, host->h_length);
   } else return(-1);
 
 if (port) sin.sin_port= htons(port);
 else
   return(-2);

 net=socket(AF_INET, SOCK_STREAM, 0);
 alarm_abort=0;alarm(timeout);
 old_alarm=signal(SIGALRM,sig_alarm);
 connect(net, (struct sockaddr *)&sin, sizeof (sin));
 alarm(0);signal(SIGALRM,old_alarm);
 if(alarm_abort)return(-3);
 return(net);
}


#define BASE64_LEN(len) (4 * (((len) + 2) / 3) +2)
/* BASE64 encoding: 3x8 bits -> 4x6 bits */
static char base64[64] = 
"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";

void base64_encode (const char *s, char *store, int length)
{
int i;
unsigned char *p = (unsigned char *)store;

	for (i = 0; i < length; i += 3){
		*p++ = base64[s[0] >> 2];
		*p++ = base64[((s[0] & 3) << 4) + (s[1] >> 4)];
		*p++ = base64[((s[1] & 0xf) << 2) + (s[2] >> 6)];
		*p++ = base64[s[2] & 0x3f];
		s += 3;
	}
	/*Pad the result*/
	if (i == length + 1) *(p - 1) = '=';
	else if (i == length + 2) *(p - 1) = *(p - 2) = '=';
	*p = '\0';
}

char *HTTPErrMsg(int code){
	switch(code){
	case 0:   return("Not indexed yet");
	case 200: return("OK");
	case 301: return("Moved Permanently");
	case 302: return("Moved Temporarily");
	case 303: return("See Other");
	case 304: return("Not Modified");
	case 300: return("Multiple Choices");
	case 305: return("Use Proxy (proxy redirect)");
	case 400: return("Bad Request");
	case 401: return("Unauthorized");
	case 402: return("Payment Required");
	case 403: return("Forbidden");
	case 404: return("Not found");
	case 405: return("Method Not Allowed");
	case 406: return("Not Acceptable");
	case 407: return("Proxy Authentication Required");
	case 408: return("Request Timeout");
	case 409: return("Conflict");
	case 410: return("Gone");
	case 411: return("Length Required");
	case 412: return("Precondition Failed");
	case 413: return("Request Entity Too Large");
	case 414: return("Request-URI Too Long");
	case 415: return("Unsupported Media Type");
	case 500: return("Internal Server Error");
	case 501: return("Not Implemented");
	case 502: return("Bad Gateway");
	case 505: return("HTTP Version Not Supported");
	case 503: return("Service Unavailable");
	case 504: return("Gateway Timeout");
	default:  return("Unknown status");
	}
}

int HTTPResponseType(int status){

	switch(status){
	case 200: return(HTTP_STATUS_OK);
	
	case 301:  /* Moved Permanently */
	case 302:  /* Moved Temporarily */
	case 303:  /* See Other */ return(HTTP_STATUS_REDIRECT);
	
	case 304:  /* Not Modified*/ return(HTTP_STATUS_NOT_MODIFIED);
	
	case 300:  /* Multiple Choices              */
	case 305:  /* Use Proxy (proxy redirect)    */
	case 400:  /* Bad Request ??? We tried ...  */
	case 401:  /* Unauthorized                  */
	case 402:  /* Payment Required              */
	case 403:  /* Forbidden                     */
	case 404:  /* Not found                     */
	case 405:  /* Method Not Allowed            */
	case 406:  /* Not Acceptable                */
	case 407:  /* Proxy Authentication Required */
	case 408:  /* Request Timeout               */
	case 409:  /* Conflict                      */
	case 410:  /* Gone                          */
	case 411:  /* Length Required               */
	case 412:  /* Precondition Failed           */
	case 413:  /* Request Entity Too Large      */
	case 414:  /* Request-URI Too Long          */
	case 415:  /* Unsupported Media Type        */
	case 500:  /* Internal Server Error         */
	case 501:  /* Not Implemented               */
	case 502:  /* Bad Gateway                   */
	case 505:  /* HTTP Version Not Supported    */ 
		return(HTTP_STATUS_DELETE);

	case 503: /* Service Unavailable */
	case 504: /* Gateway Timeout */ 
		return(HTTP_STATUS_RETRY);

	default: return(HTTP_STATUS_UNKNOWN);
	}
}

int HTTPGet(char *host,int port,char *header,
char *dist,int maxsize,int read_timeout){
int fd,i,size;
char *s;
void *old_alarm;

	/* Connect to HTTP or PROXY server */
	fd=open_host(host,port,read_timeout);	/* Connect           */
	if(fd<0)return(fd);			/* Could not connect */

	alarm_abort=0;alarm(read_timeout);
	old_alarm=signal(SIGALRM,sig_alarm);
	write(fd,header,strlen(header));/* Send HTTP request */
	size=0;s=dist;			/* Retrieve response */
	while((i=read(fd,s,maxsize-1-size))){
		size+=i;s+=i;
		if(alarm_abort)break;
		else signal(SIGALRM,sig_alarm);
	}
	close(fd);/* Close HTTP connection */
	alarm(0);signal(SIGALRM,old_alarm);
	if(alarm_abort)return(-3);
	return(size);
}

int NEWSGet(char *host,int port,char *header,
	char *dist,int maxsize,int read_timeout){
int fd,size,status;
char *s,*c,*e;
void *old_alarm;
char str[STRSIZ];
FILE *f,*f1;
char filename[STRSIZ];
char command[32]="";
char proto[32]="";
int a=0,b=0,from=0,to=0,headers=0;
char *email=NULL;
char *msg_id=NULL;
char *subj=NULL;
int has_content=0;
int has_if_modified=0;


	sscanf(header,"%s%s%s",command,str,proto);
	if(str[0]=='/')strcpy(filename,str+1);
	else	strcpy(filename,str);
	s=strtok(header,"\r\n");
	while(s){
		if(!STRNCASECMP(s,"If-Modified-Since: "))
			has_if_modified=1;
		s=strtok(NULL,"\r\n");
	}

        /* Connect to NEWS server */
	fd=open_host(host,port,read_timeout);	/* Connect           */
	if(fd<0)return(fd);			/* Could not connect */

	alarm_abort=0;alarm(read_timeout);
	old_alarm=signal(SIGALRM,sig_alarm);
	size=0;s=dist;
	f=fdopen(fd,"r");f1=fdopen(fd,"w");
	fgets(str,sizeof(str),f); /* Read hello string */

	/* All news groups on the server */
	if((!strcmp(filename,"/"))||(!strcmp(filename,""))){
		fprintf(f1,"list\r\n");fflush(f1);
		fgets(str,sizeof(str),f);
		sprintf(dist,"HTTP/1.0 200 OK\r\nContent-Type: text/html\r\n\r\n<HTML><BODY>\n");
		while(fgets(str,sizeof(str),f)){
			if(str[0]=='.')break;
			if((c=strchr(str,' ')))*c=0;
			sprintf(STREND(dist),"<A HREF=\"news://%s/%s\"></A>\n",host,str);
			if(alarm_abort)break;
			else signal(SIGALRM,sig_alarm);
		}
		sprintf(STREND(dist),"</BODY></HTML>\n");
	}else /* List of the articles in one news group */
	if(!strchr(filename,'@')){
		fprintf(f1,"group %s\r\n",filename);fflush(f1);
		fgets(str,sizeof(str),f);
		sscanf(str,"%d%d%d%d",&a,&b,&from,&to);
		fprintf(f1,"xover %d-%d\r\n",from,to);fflush(f1);
		fgets(str,sizeof(str),f);
		sprintf(dist,"HTTP/1.0 200 OK\r\nContent-Type: text/html\r\n\r\n<HTML><BODY>\n");
		while(fgets(str,sizeof(str),f)){
			if(str[0]=='.')break;
			if(alarm_abort)break;
			else signal(SIGALRM,sig_alarm);
			if((e=strtok(str,"\t\r\n")))
			if((subj=strtok(NULL,"\t\r\n")))
			if((email=strtok(NULL,"\t\r\n")))
			if((e=strtok(NULL,"\t\r\n")))
			if((msg_id=strtok(NULL,"\t\r\n")));
			str[0]=0;
			if(msg_id)sscanf(msg_id,"<%[^>]s>",str);
			if(str[0])sprintf(STREND(dist),"<A HREF=\"news://%s/%s\"></A>\n",host,str);
		}
		sprintf(STREND(dist),"</BODY></HTML>\n");
	}else{ /* One article */
		if(has_if_modified){
			fprintf(f1,"stat <%s>\r\n",filename);fflush(f1);
			fgets(str,sizeof(str),f);sscanf(str,"%d",&a);
			if(a==223){
				status=304; /* Not modified */
				sprintf(dist,"HTTP/1.0 %d %s\r\n\r\n",status,HTTPErrMsg(status));
			}else{
				status=404; /* Not found */
				sprintf(dist,"HTTP/1.0 %d %s\r\n\r\n",status,HTTPErrMsg(status));
			}
		}else{
			fprintf(f1,"article <%s>\r\n",filename);fflush(f1);
			fgets(str,sizeof(str),f);sscanf(str,"%d",&a);
			if(a==220){
				headers=1;has_content=0;
				sprintf(dist,"HTTP/1.0 200 OK\r\n");
				while(fgets(str,sizeof(str),f)){
					if(str[0]=='.')break;
					if(alarm_abort)break;
					else signal(SIGALRM,sig_alarm);
					if(headers){
						if(!STRNCASECMP(str,"Content-Type:"))has_content=1;
						if((!strcmp(str,"\r\n"))||(!strcmp(str,"\n"))){
							headers=0;
							if(!has_content)strcat(dist,"Content-Type: text/plain\r\n");
						}
					}
					if((strlen(dist)+strlen(str))<maxsize)
						strcat(dist,str);
				}
			}else{
				status=404;
				sprintf(dist,"HTTP/1.0 %d %s\r\n\r\n",status,HTTPErrMsg(status));
			}
		}
	}
	fprintf(f1,"quit\r\n");fflush(f1);	
	fclose(f);/* Close NEWS connection */
	fclose(f1);
	alarm(0);signal(SIGALRM,old_alarm);
	if(alarm_abort)return(-3);
	return(strlen(dist));
}


int NNTPGet(char *host,int port,char *header,
	char *dist,int maxsize,int read_timeout){
int fd,size,status;
char *s,*c,*e;
void *old_alarm;
char str[STRSIZ];
FILE *f,*f1;
char filename[STRSIZ];
char command[32]="";
char proto[32]="";
int a=0,b=0,from=0,to=0,headers=0;
int has_content=0;
int has_if_modified=0;


	sscanf(header,"%s%s%s",command,str,proto);
	printf("str='%s'\n",str);
	if(str[0]=='/')strcpy(filename,str+1);
	else	strcpy(filename,str);

	s=strtok(header,"\r\n");
	while(s){
		if(!STRNCASECMP(s,"If-Modified-Since: "))has_if_modified=1;
		s=strtok(NULL,"\r\n");
	}

	/* Connect to NNTP server */
	fd=open_host(host,port,read_timeout);	/* Connect           */
	if(fd<0)return(fd);			/* Could not connect */

	alarm_abort=0;alarm(read_timeout);
	old_alarm=signal(SIGALRM,sig_alarm);
	size=0;s=dist;
	f=fdopen(fd,"r");f1=fdopen(fd,"w");
	fgets(str,sizeof(str),f); /* Read hello string */

	/* All news groups on the server */
	if((!strcmp(filename,"/"))||(!strcmp(filename,""))){
		fprintf(f1,"list\r\n");fflush(f1);
		fgets(str,sizeof(str),f);
		sprintf(dist,"HTTP/1.0 200 OK\r\nContent-Type: text/html\r\n\r\n<HTML><BODY>\n");
		while(fgets(str,sizeof(str),f)){
			if(str[0]=='.')break;
			if((c=strchr(str,' ')))*c=0;
			sprintf(STREND(dist),"<A HREF=\"nntp://%s/%s\"></A>\n",host,str);
			if(alarm_abort)break;
			else signal(SIGALRM,sig_alarm);
		}
		sprintf(STREND(dist),"</BODY></HTML>\n");
	}else /* List of the articles in one news group */
	if(!strchr(filename,'@')){
		fprintf(f1,"group %s\r\n",filename);fflush(f1);
		fgets(str,sizeof(str),f);
		sscanf(str,"%d%d%d%d",&a,&b,&from,&to);
		fprintf(f1,"xover %d-%d\r\n",from,to);fflush(f1);
		fgets(str,sizeof(str),f);
		sprintf(dist,"HTTP/1.0 200 OK\r\nContent-Type: text/html\r\n\r\n<HTML><BODY>\n");
		while(fgets(str,sizeof(str),f)){
			if(str[0]=='.')break;
			if(alarm_abort)break;
			else signal(SIGALRM,sig_alarm);
			if((e=strtok(str,"\t\r\n")))
				sprintf(STREND(dist),
				"<A HREF=\"nntp://%s/%s/%s\"></A>\n",
				host,filename,e);
		}
		sprintf(STREND(dist),"</BODY></HTML>\n");
	}else{ /* One article */
		if(has_if_modified){
			fprintf(f1,"stat <%s>\r\n",filename);fflush(f1);
			fgets(str,sizeof(str),f);sscanf(str,"%d",&a);
			if(a==223){
				status=304; /* Not modified */
				sprintf(dist,"HTTP/1.0 %d %s\r\n\r\n",status,HTTPErrMsg(status));
			}else{
				status=404; /* Not found */
				sprintf(dist,"HTTP/1.0 %d %s\r\n\r\n",status,HTTPErrMsg(status));
			}
		}else{
			fprintf(f1,"article <%s>\r\n",filename);fflush(f1);
			fgets(str,sizeof(str),f);sscanf(str,"%d",&a);
			if(a==220){
				headers=1;has_content=0;
				sprintf(dist,"HTTP/1.0 200 OK\r\n");
				while(fgets(str,sizeof(str),f)){
					if(str[0]=='.')break;
					if(alarm_abort)break;
					else signal(SIGALRM,sig_alarm);
					if(headers){
						if(!STRNCASECMP(str,"Content-Type:"))has_content=1;
						if((!strcmp(str,"\r\n"))||(!strcmp(str,"\n"))){
							headers=0;
							if(!has_content)strcat(dist,"Content-Type: text/plain\r\n");
						}
					}
					if((strlen(dist)+strlen(str))<maxsize)
						strcat(dist,str);
				}
			}else{
				status=404;
				sprintf(dist,"HTTP/1.0 %d %s\r\n\r\n",status,HTTPErrMsg(status));
			}
		}
	}
	fprintf(f1,"quit\r\n");fflush(f1);	
	fclose(f);/* Close NNTP connection */
	fclose(f1);
	alarm(0);signal(SIGALRM,old_alarm);
	if(alarm_abort)return(-3);
	return(strlen(dist));
}


int FILEGet(char *header,char *dist,int maxsize){
int fd,size,l,status=0,is_dir;
struct stat sb,sb1;
DIR * dir;
struct dirent *rec;
char filename[STRSIZ];
char newfilename[STRSIZ];
char command[32]="";
char proto[32]="";
char *content_type=NULL;

	sscanf(header,"%s%s%s",command,filename,proto);

	if(stat(filename,&sb)){
		/*
		status=500;
		sprintf(dist,"HTTP/1.0 %d %s\r\n\r\n",status,HTTPErrMsg(status));
		return(strlen(dist));
		*/
	}

	if((sb.st_mode&S_IFDIR)&&(filename[strlen(filename)-1]!='/')){
		status=301;
		sprintf(dist,"HTTP/1.0 %d %s\r\nLocation: file:%s/\r\n\r\n",status,HTTPErrMsg(status),filename);
		return(strlen(dist));
	}
	if(sb.st_mode&S_IFDIR){
		if((dir=opendir(filename))){
			strcpy(dist,"HTTP/1.0 200 OK\r\nContent-type: text/html\r\n\r\n<HTML><BODY>\n");
			while((rec = readdir (dir))){
				/*
				This does not work on Solaris
				is_dir=(rec->d_type==DT_DIR);
				*/
				sprintf(newfilename,"%s%s",filename,rec->d_name);
				if(stat(newfilename,&sb1)){
					 /* Error handler should be here !!! */
				}
				is_dir=((sb1.st_mode&S_IFDIR)>0);
				sprintf(STREND(dist),"<A HREF=\"%s%s\">%s%s</A>\n",
					rec->d_name,is_dir?"/":"",
					rec->d_name,is_dir?"/":"");
			}
			closedir(dir);
			strcpy(STREND(dist),"</BODY><HTML>\n");
			return(strlen(dist));
		}else{
			switch(errno){
				case ENOENT: status=404;break; /* Not found */
				case EACCES: status=403;break; /* Forbidden*/
				default: status=1;
			}
			sprintf(dist,"HTTP/1.0 %d %s\r\n\r\n",status,HTTPErrMsg(status));
			return(strlen(dist));
		}
	}else{
		if((fd=open(filename,O_RDONLY))<0){
			switch(errno){
				case ENOENT: status=404;break;
				case EACCES: status=403;break;
				default: status=1;
			}
			sprintf(dist,"HTTP/1.0 %d %s\r\n\r\n",status,HTTPErrMsg(status));
			return(strlen(dist));
		}
		sprintf(dist,"HTTP/1.0 200 OK\r\n");
		if((content_type=ContentType(filename)))
			sprintf(STREND(dist),"Content-Type: %s\r\n",content_type);
		sprintf(STREND(dist),"\r\n");
		l=strlen(dist);
		size=read(fd,dist+l,maxsize-l)+l;
		close(fd);
		return(size);
	}
}
